package edu.illinois.reassert;

import java.io.File;
import java.io.IOException;
import java.util.List;
import java.util.ListIterator;
import java.util.Set;

import org.apache.commons.io.FileUtils;

import spoon.reflect.code.CtBlock;
import spoon.reflect.cu.CompilationUnit;
import spoon.reflect.cu.SourceCodeFragment;
import spoon.reflect.cu.SourcePosition;
import spoon.reflect.declaration.CtElement;
import spoon.reflect.declaration.CtExecutable;
import spoon.reflect.declaration.CtPackage;
import spoon.reflect.declaration.CtSimpleType;
import spoon.reflect.declaration.CtVariable;
import spoon.reflect.reference.CtExecutableReference;
import spoon.reflect.reference.CtTypeReference;
import spoon.reflect.visitor.CtScanner;

/**
 * Represents a {@link FixResult} in which code has changed
 * 
 * @author <a href="http://www.cs.uiuc.edu/homes/bdaniel3/">Brett Daniel</a>
 */
public class CodeFixResult extends FixResult {

	/**
	 * @deprecated Needed for the mutation evaluation
	 */
	public static boolean SHIFT_LINE_NUMBERS = true;
	
	private CtElement fixedElement;
	private SimpleSpoonLoader loader;
	private CompilationUnit fixedCU;
	private SourceCodeFragment fragment;
	
	public CodeFixResult(CtElement fixedElement) {
		if (fixedElement.getParent() == null) {
			throw new IllegalArgumentException("Fixed element parent is not set");
		}
		SourcePosition fixPosition = fixedElement.getPosition();
		if (fixPosition == null) {
			throw new IllegalArgumentException("Fixed element position is not set");
		}
		// TODO: A lot of logic here in the constructor. May be good to move it
		// elsewhere, but it is necessary to apply the fragment to the compilation
		// unit immediately.
		this.fixedElement = fixedElement;
		this.fixedCU = fixPosition.getCompilationUnit();		
		fragment = buildFragment(fixedCU, fixedElement);		
		addFragment(fixedCU, fragment);
		if (SHIFT_LINE_NUMBERS) {
			shiftLineNumbers(fixedCU, fragment, fixPosition);
		}
	}

	/**
	 * If this fix adds lines to the compilation unit, then adjust
	 * the positions of all later elements.  
	 */
	private void shiftLineNumbers(
			CompilationUnit cu, 
			SourceCodeFragment frag, 
			SourcePosition fixPosition) {
		final int startLine = fixPosition.getLine();
		final int endLine = fixPosition.getEndLine();
		final int originalLines = endLine - startLine + 1;
		final int offset = frag.code.split("\n").length - originalLines;
		if (offset != 0) {
			CtScanner lineScanner = new CtScanner() {
				@Override
				protected void enter(CtElement e) {
					final SourcePosition position = e.getPosition();
					if (position != null) {
						int startOffset = 0;
						int endOffset = 0;
						if (startLine < position.getLine()) {
							// fragment comes before element. shift start line
							startOffset = offset;
						}
						if (endLine <= position.getEndLine()) {
							// fragment comes before or inside element. shift end line
							endOffset = offset;
						}
						if (startOffset != 0 || endOffset != 0) {
							e.setPosition(new OffsetPosition(position, startOffset, endOffset));
						}
					}
				}
			};
			for (CtSimpleType<?> type : cu.getDeclaredTypes()) {
				type.accept(lineScanner);
			}
		}
	}
	
	private void addFragment(CompilationUnit cu, SourceCodeFragment frag) {
		List<SourceCodeFragment> frags = cu.getSourceCodeFraments();
		if (frags != null) {
			// HACK: Spoon's addSourceCodeFragment does not replace overlapping
			// fragments.  Since ReAssert does that a lot, we need to do it here.
			ListIterator<SourceCodeFragment> fragIter = frags.listIterator();
			while (fragIter.hasNext()) {
				SourceCodeFragment existing = fragIter.next();
				if (areOverlapping(existing, frag)) {
					// replace overlapping fragment
					fragIter.remove();
					fragIter.add(frag);
					return;
				}
			}
		}
		cu.addSourceCodeFragment(frag);
	}

	private boolean areOverlapping(
			SourceCodeFragment oldFrag,
			SourceCodeFragment newFrag) {
		int newStart = newFrag.position;
		int newEnd = newStart + newFrag.replacementLength;
		int oldStart = oldFrag.position;
		int oldEnd = oldStart + oldFrag.replacementLength;
		return newStart <= oldStart	&& oldEnd <= newEnd;
	}

	private SourceCodeFragment buildFragment(
			CompilationUnit cu, 
			final CtElement fixedElement) {
		CtTypeReference<?> junit3 = fixedElement.getFactory().Type().createReference(junit.framework.Assert.class);
		CtTypeReference<?> junit4 = fixedElement.getFactory().Type().createReference(org.junit.Assert.class);

		ImportScanner importer = new ImportScanner();
		importer.ignore(fixedElement);
		Set<CtTypeReference<?>> imports = importer.makeImports(cu);
		imports.add(junit4);
		imports.add(junit3);		

		StaticImportScanner siImporter = new StaticImportScanner();
		siImporter.importStaticFrom(junit3);
		siImporter.importStaticFrom(junit4);
		Set<CtExecutableReference<?>> staticImports = siImporter.makeStaticImports(cu);
		
		ReAssertPrettyPrinter pp = new ReAssertPrettyPrinter(
				fixedElement.getFactory().getEnvironment());
		pp.setTabCount(getNestingDepth(fixedElement));
		pp.addImports(imports);
		pp.addStaticImports(staticImports);
		pp.scan(fixedElement);
		String code = pp.toString();
		
		SourcePosition fixPosition = fixedElement.getPosition();
		int start = fixPosition.getSourceStart();
		int end = fixPosition.getSourceEnd();
		
		int leftChars = 0;
		int rightChars = 1;
		// The character offsets stored in the fixed element do not 
		// always correspond to the full text of the fixed element.
		if (fixedElement instanceof CtBlock) {
			// a block counts from inside the braces,
			// so when replacing the fragment, we need to include the
			// brace characters. 
			if (fixedElement.getParent() instanceof CtExecutable) {
				leftChars = 1;
			}
			rightChars = 2			;
		} 
		else if (fixedElement instanceof CtVariable) {
			// a variable does not count the type
			pp.reset();
			pp.scan(((CtVariable<?>) fixedElement).getType());
			leftChars = pp.toString().length() + 1;
		}
		// might need to add other element types
		
		SourceCodeFragment frag = new SourceCodeFragment();
		frag.position = start - leftChars;
		frag.replacementLength = end - start + leftChars + rightChars;			
		frag.code = code;
		return frag;
	}

	private int getNestingDepth(CtElement elem) {
		if (elem == null || elem instanceof CtPackage) {
			// package has nesting depth -1 (top-level class has 0), 
			// subtract 1 again because we ignore the target type
			return -2; 
		}
		return 1 + getNestingDepth(elem.getParent());
	}

	public CtElement getFixedElement() {
		return fixedElement;
	}

	public void setLoader(SimpleSpoonLoader loader) {
		this.loader = loader;
	}
	
	@Override
	public File save(File outDir, String fixFileSuffix) throws IOException {
		File fixFile = new File(outDir, fixedCU.getFile().getPath() + fixFileSuffix);
		loader.output(fixedCU, fixFile);
		return fixFile;
	}

	@Override
	public File saveSnippet(File outDir, String snippetFileSuffix) throws IOException {
		String fileName = String.format("%s.%d.%d%s", 
				fixedCU.getFile().getPath(),
				fragment.position,
				fragment.replacementLength,
				snippetFileSuffix);
		File fixFile = new File(outDir, fileName);
		FileUtils.writeStringToFile(fixFile, fragment.code);
		return fixFile;
	}
	
}
