package edu.illinois.reassert;

import java.io.File;
import java.lang.reflect.Method;
import java.util.Collection;
import java.util.List;
import java.util.ListIterator;

import spoon.reflect.code.CtBlock;
import spoon.reflect.code.CtStatement;
import spoon.reflect.cu.CompilationUnit;
import spoon.reflect.cu.SourcePosition;
import spoon.reflect.declaration.CtElement;
import spoon.reflect.declaration.CtSimpleType;
import spoon.reflect.visitor.Query;
import spoon.reflect.visitor.filter.AbstractFilter;

/**
 * Default abstract implementation of {@link FixStrategy} 
 * 
 * @author <a href="http://www.cs.uiuc.edu/homes/bdaniel3/">Brett Daniel</a>
 */
public abstract class FixStrategyBase implements FixStrategy {

	private final Factory factory;
	
	public FixStrategyBase(Factory factory) {
		this.factory = factory;
	}
	
	public Factory getFactory() {
		return factory;
	}
	
	/**
	 * @return the {@link AssertFactory} for the given test method
	 */
	protected AssertFactory getAssertFactory(Method testMethod) {
		return getFactory().Assert(testMethod);
	}

	
	/**
	 * {@inheritDoc}
	 */
	@Override
	public Collection<String> getMethodsToInstrument() {
		return null;
	}
	
	/**
	 * Replaces the old statement in its surrounding block with the new statement
	 */
	protected void replaceStatement(CtStatement oldStatement, CtStatement newStatement) {
		CtBlock<?> containingBlock = oldStatement.getParent(CtBlock.class);
		if (containingBlock == null) {
			// Probably indicates that a previous fix forgot to set parents correctly
			throw new RuntimeException("Unable to replace failing statement.");
		}
		ListIterator<CtStatement> statementIter = 
			containingBlock.getStatements().listIterator();
		while (statementIter.hasNext()) {
			CtStatement stmt = statementIter.next();
			if (stmt == oldStatement) {
				statementIter.remove();
				statementIter.add(newStatement);				
				break;
			}
		}
		setParents(oldStatement.getParent(), newStatement);
		newStatement.setPosition(oldStatement.getPosition());
	}
	
	/**
	 * Sets all parents for elements in the given tree
	 * @param parent the parent of the root
	 * @param root the element from which to start traversing
	 */
	protected void setParents(CtElement parent, CtElement root) {
		new ParentVisitor(parent).scan(root);
	}

	/**
	 * @return the file containing the given location
	 */
	protected File findFailingFile(StackTraceElement failureLocation) {
		String packagePath = failureLocation.getClassName();
		int packIndex = packagePath.lastIndexOf('.');
		if (packIndex > -1) {
			packagePath = packagePath.substring(0, packIndex);
			packagePath = packagePath.replace('.', File.separatorChar);
		}
		else {
			packagePath = ""; // root dir
		}
		return new File(packagePath, failureLocation.getFileName());
	}

	/**
	 * @return the {@link CompilationUnit} containing the given location or null
	 */
	protected CompilationUnit findFailingCompilationUnit(StackTraceElement location) {
		File failingFile = findFailingFile(location);
		return getFactory().CompilationUnit().find(failingFile.getPath());
	}

	/**
	 * @return the "innermost" element of the given type containing the given location or null
	 */
	protected <T extends CtElement> T findFailingElement(
			Class<T> elementType,
			final StackTraceElement location) {
		CompilationUnit cu = findFailingCompilationUnit(location);
		if (cu == null) {
			return null;
		}
		for (CtSimpleType<?> type : cu.getDeclaredTypes()) {
			List<T> elements = 
				Query.getElements(type, new AbstractFilter<T>(elementType) {
					@Override
					public boolean matches(T element) {
						return elementContainsLocation(element, location);
					}
				});
			if (elements.size() > 0) {
				// return "innermost" element
				return elements.get(elements.size() - 1);
			}
		}
		return null;
	}
	
	protected boolean elementContainsLocation(CtElement element, StackTraceElement assertLocation) {
		return elementContainsLine(element, assertLocation.getLineNumber());
	}

	protected boolean elementContainsLine(CtElement element, int lineNumber) {
		SourcePosition position = element.getPosition();
		if (position == null) {
			return false;
		}
		return position.getLine() <= lineNumber 
			&& lineNumber <= position.getEndLine();
	}

}
