package edu.illinois.reassert.assertfixer;

import java.lang.reflect.Method;
import java.util.List;
import java.util.ListIterator;

import spoon.reflect.code.CtBlock;
import spoon.reflect.code.CtLocalVariable;
import spoon.reflect.code.CtStatement;
import spoon.reflect.code.CtTry;
import edu.illinois.reassert.CodeFixResult;
import edu.illinois.reassert.Factory;
import edu.illinois.reassert.FixResult;
import edu.illinois.reassert.FixStrategyBase;
import edu.illinois.reassert.ReAssertPrettyPrinter;
import edu.illinois.reassert.RecordedAssertFailure;
import edu.illinois.reassert.UnfixableException;

/**
 * Strategy that removes any failing assertions introduced by 
 * {@link AssertEqualsExpandAccessorsFixer}.  Prevents nondeterministic
 * assertions from causing a test to fail.  
 * 
 * @author <a href="http://www.cs.uiuc.edu/homes/bdaniel3/">Brett Daniel</a>
 */
public class RemoveNondeterministicFixer extends FixStrategyBase {

	public RemoveNondeterministicFixer(Factory factory) {
		super(factory);
	}

	@Override
	public FixResult fix(
			Method testMethod, 
			Throwable failureException)
			throws UnfixableException {		
		if (!(failureException instanceof RecordedAssertFailure)) {
			return null;
		}
		RecordedAssertFailure record = (RecordedAssertFailure) failureException;
		StackTraceElement location = record.getStackTrace()[1]; // one above instrumented
		
		CtBlock<?> failingBlock = findFailingElement(CtBlock.class, location);
		int startLine = failingBlock.getPosition().getLine();
		int lines = removeFailingAssertion(startLine, location.getLineNumber(), failingBlock);
		if (lines == -1) {
			if (isEmpty(failingBlock)) {
				throw new UnfixableException("Object has no deterministic accessors");
			}
			return new CodeFixResult(failingBlock);
		}
		return null;
	}

	/**
	 * @return true if the block contains no statements or only a dummy variable
	 */
	private boolean isEmpty(CtBlock<?> block) {
		List<CtStatement> statements = block.getStatements();
		return 
			statements.size() == 0
				|| ( statements.size() == 1 
						&& statements.get(0) instanceof CtLocalVariable);
	}

	/**
	 * Remove the statement at the given line in the block
	 * @return the current line or -1 if a line was removed
	 */
	private int removeFailingAssertion(
			int curLine,
			int failingLine, 
			CtBlock<?> failingBlock) {
		ListIterator<CtStatement> stmtIter = 
			failingBlock.getStatements().listIterator();
		while (stmtIter.hasNext() && curLine > -1) {
			CtStatement stmt = stmtIter.next();
			curLine++;
			if (stmt instanceof CtBlock) {
				curLine = removeFailingAssertion(
						curLine, failingLine, (CtBlock<?>) stmt);
				if (curLine > -1) {
					curLine++; // +1 for closing brace
				}
				else if (isEmpty((CtBlock<?>) stmt)) {
					stmtIter.remove();
				}
			}
			else if (stmt instanceof CtTry) {
				curLine = removeFailingAssertion(
						curLine, failingLine, ((CtTry) stmt).getBody());
				assert curLine > -1; // shouldn't remove anything from try block
				curLine++; // +1 for closing brace 
				curLine = removeFailingAssertion(
						curLine, failingLine, ((CtTry) stmt).getCatchers().get(0).getBody());
				if (curLine > -1) {
					curLine++; // +1 for closing brace
				}
			}
			else if (curLine == failingLine
					&& stmt.getPosition() == null // null because it was added by ReAssert
					) {
				stmtIter.remove();
				return -1;
			}
			else {
				curLine += lineLength(stmt) - 1;
			}
		}
		return curLine;
	}

	private int lineLength(CtStatement stmt) {
		ReAssertPrettyPrinter pp = new ReAssertPrettyPrinter(getFactory().getEnvironment());
		pp.reset();
		pp.scan(stmt);
		return pp.getResult().toString().split("\n").length;
	}

}
